#Licensed Materials - Property of IBM
#IBM SPSS Products: Statistics General
#(c) Copyright IBM Corp. 2011
#US Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

"""This module contains a helper function that displays the category structure of a pivot table.
It is intended to be used with the spssaux function getValuesFromXmlWorkspace."""

__author__  =  'JKP, IBM SPSS'
__version__ =  '1.0.0'
version = __version__

# history
# 22-Feb-2010 original version

import spss, spssaux
# Cannot use cElementTree because of api differences
from xml.etree import ElementTree
import re

try:
    _("---")
except:
    def _(msg):
        return msg
    
v19okay = spssaux.getSpssMajorVersion() >= 19

# Version 18 has a bug related to inserting Unicode text in a pivot table data cell.
# The following code monkey patches the problematic method just for V18 and earlier

if not v19okay:
    from spss import CellText, Dimension
    import time, datetime
    
    def SimplePivotTable(self,rowdim="",rowlabels=[],coldim="",collabels=[],cells=None):
        """Monkey Patch for unicode-in-cells bug for V18 and earlier"""

        error.Reset()

        try:
            # If cells is None, neither rowlabels nor collabels is allowed.
            if cells is None:
                if len(rowlabels) > 0:
                    error.SetErrorCode(1032)
                    if error.IsError():
                        raise SpssError,error
                elif len(collabels) > 0:
                    error.SetErrorCode(1032)
                    if error.IsError():
                        raise SpssError,error

            # Make a local copy. We don't want to change the original labels.
            tmpRowLabels = list(rowlabels)
            tmpColLabels = list(collabels)
        except TypeError:
            error.SetErrorCode(1004)
            if error.IsError():
                raise SpssError,error

        # Check the structure of cells.
        nRows = 0
        nCols = 0

        # None or empty cells is okay at this point.
        if cells is not None:
            nRows = len(cells)
            if nRows > 0:
                if not isinstance(cells[0],str):
                    try:
                        #check if cells[0] is iterable.
                        nCols = len([(i, x) for (i,x) in enumerate(cells[0])])
                    except TypeError:
                        nCols = 1
                else:
                    nCols = 1

        if tmpRowLabels <> [] and tmpColLabels <> []:
            nRows = len(tmpRowLabels)
            nCols = len(tmpColLabels)
        elif tmpRowLabels <> []:
            nRows = len(tmpRowLabels)
            # If there are no labels for the column dimension, the length of the first cells item is used.
            tmpColLabels.extend(["col"+str(x) for x in range(nCols)])
        elif tmpColLabels <> []:
            nCols = len(tmpColLabels)
            # If there are no labels for the row dimension, the number of rows in Cells is used.
            tmpRowLabels.extend(["row"+str(x) for x in range(nRows)])
        else:
            tmpRowLabels.extend(["row"+str(x) for x in range(nRows)])
            tmpColLabels.extend(["col"+str(x) for x in range(nCols)])

        tmpRowLabels = map(CellText._CellText__ToCellText,tmpRowLabels)
        tmpColLabels = map(CellText._CellText__ToCellText,tmpColLabels)

        tmpCells = []

        # cells must match the label structure if the labels are given.
        if nRows > 0 and nCols > 0:
            try:
                # Check cells length and if cells can be indexed as cells[i][j] or cells[i].
                if nCols > 1:
                    try:
                        x = []
                        for c in cells:
                            if isinstance(c, (tuple,list)):
                                x.append(len(c))
                            else:
                                x.append(1)
                        maxlen = max(x)
                    except TypeError:
                        maxlen = 1
                    if ( 1 == maxlen ):
                        assert (len(cells) == nCols * nRows)
                        tmpCells = [cells[x*nCols + y] for x in range(nRows) for y in range(nCols)]
                    else:
                        assert(maxlen == nCols)
                        assert(len(cells) == nRows)
                        tmpCells = [cells[x][y] for x in range(nRows) for y in range(nCols)]
                else:
                    assert(len(cells) == nCols * nRows)
                    tmpCells = [cells[x*nCols + y] for x in range(nRows) for y in range(nCols)]
            except:
                error.SetErrorCode(1032)
                if error.IsError():
                    raise SpssError, error

        # Check if cells[i][j] or cells[i] is scalar (such as sequence).
        for x in tmpCells:
            ###if not isinstance(x,(str, time.struct_time, datetime.datetime, datetime.date)):
            if not isinstance(x,(basestring, time.struct_time, datetime.datetime, datetime.date)):
                try:
                    [(i, x) for (i,x) in enumerate(x)]
                    error.SetErrorCode(1032)
                    if error.IsError():
                        raise SpssError, error
                except TypeError:
                    pass

        tmpCells = map(CellText._CellText__ToCellText,tmpCells)

        # If dimension is empty, the dimension label is hidden.
        if rowdim == "":
            rowdim = self.Append(Dimension.Place.row,"rowdim",True,False)
        else:
            rowdim = self.Append(Dimension.Place.row,rowdim,False,False)
        if coldim == "":
            coldim = self.Append(Dimension.Place.column,"coldim",True,False)
        else:
            coldim = self.Append(Dimension.Place.column,coldim,False,False)

        if tmpCells <> []:
            categories = [(row,col) for row in tmpRowLabels for col in tmpColLabels]
            for (i,cats) in enumerate(categories):
                self[cats] = tmpCells[i]

    # monkey patch BasePivotTable class
    import spss.errMsg
    error = spss.errMsg.errCode()
    spss.BasePivotTable.SimplePivotTable = SimplePivotTable


def _safeval(val, quot='"'):
    "return safe value for quoting with quot, which may be single or double quote or blank"
    return quot == " " and val or val.replace(quot, quot+quot)


def displayTableStructure(cmd, omsid, subtype):
    """Display a table showing the row and column structure of a particular pivot table
    
    cmd is the syntax to generate the table
    omsid is the OMS command identifier for the command
    subtype is the OMS identifier for the table in question
    
    The main purpose of this function is to make it easier to construct the arguments
    for getValuesFromXmlWorkspace.  The table structure is flattened in the same
    way that it would be used by that function.  Group structure is not shown.
    
    If the command generated more than one pivot table, the first one is analyzed.
    
    The command output will be visible.
    
    Tables with layers are not handled by this function"""
    
    #debugging...
    #try:
        #import wingdbstub
        #if wingdbstub.debugger != None:
            #import time
            #wingdbstub.debugger.StopDebug()
            #time.sleep(2)
            #wingdbstub.debugger.StartDebug()
        #import thread
        #wingdbstub.debugger.SetDebugThreads({thread.get_ident(): 1}, default_policy=0)
        ## for V19 use
        ##    ###SpssClient._heartBeat(False)
    #except:
        #pass
    
    tag, err = spssaux.createXmlOutput(cmd, omsid, subtype, visible=True)
    if not err == 0:
        raise ValueError(_("Command specified to displayTableStructure failed"))
    
    xml = spss.GetXmlUtf16(tag)
    spss.DeleteXPathHandle(tag)
    
    # clean up the XML
    # convert from UTF-16 and modify the encoding declaration accordingly.
    # Remove namespace declarations
    
    xml = xml.encode("UTF8") 
    xml = re.sub(r'''encoding="UTF-16"''', '''encoding="UTF-8"''', xml, 1)
    xml = re.sub(r"""{http://xml.spss.com/spss/oms}""", "", xml)
    xml = re.sub(r'''xmlns=".*?"''',"", xml)
    
    tree = ElementTree.fromstring(xml)
    pts = [pt for pt in tree.getiterator(tag="pivotTable") if pt.attrib["subType"] == subtype]
    if not pts:
        raise ValueError(_("Pivot table not found. omsid: %(omsid)s, subtype: %(subtype)s") % locals())
    pt = pts[-1]
    rowcats = []
    colcats = []
    
    # go through the table and find all the row and column categories
    # The main iterator passes through every element without indicating hierarchy
    # So as columns are processed, their element types are changes so that they
    # will be ignored when processing row entries.
    
    for el in pt.getiterator():
        if el.tag == "dimension":
            axis = el.attrib['axis']
            if axis == "layer":
                raise ValueError(_("Tables with layers are not supported by this function"))
            elif axis == "column":
                for celt in el.getiterator(tag="category"):
                    colcats.append(celt.attrib)
                    celt.tag = "ignore"
        elif el.tag == "category":
            rowcats.append(el.attrib)
    
    # display the result as a pivot table
    ptrows = Ptrows(rowcats, colcats, subtype, len(pts))
    ptrows.addrows()
    
    
class Ptrows(object):
    """Construct a pivot table from the xml scan results"""
    def __init__(self, rowcats, colcats, subtype, tablecount):
        """rowcats and colcats are lists of attribute dictionaries for the category
        elements.
        subtype is the table subtype (only used for labelling)
        tablecount is the number of tables found"""
        
        attributesFromDict(locals())
        self.ptattrs = set()
        
        # dedup colcats, since column structure is repeated for every row
        newlis = []
        for item in self.colcats:
            if not item in newlis:
                newlis.append(item)
        self.colcats = newlis
        
        self.allkeys = [item.keys() for item in self.rowcats + self.colcats]
        for item in self.allkeys:
            for k in item:
                self.ptattrs.add(k)
        
        #get alphabetical list of attribute names.  These will be the columns of the table
        self.columns =  sorted(self.ptattrs)
        
        caption = _("""Labels are flattened, and groups are not shown.  Table does not indicate the row and column combinations that actually occur
Cell attributes can be one or more of number, string, varName, text, text_eng, duration, or date.""")
        
        if self.tablecount > 1:
            caption += _("\nMultiple tables were found.  This table is based on the last one found")
        
        self.c = NonProcPivotTable(omssubtype="PIVOTTABLESTRUCTURE", 
            outlinetitle=_("Pivot Table Structure"),
            tabletitle=_("Row and Column Label Category Attributes for Table Type: %s") % self.subtype,
            caption = caption,
            columnlabels=["Dimension"] + self.columns, procname=_("Table Structure"))
    
    def addrows(self):
        """add all rows to the table and generate"""
        
        
        for item in self.rowcats:
            row = ["row"]
            for col in self.columns:
                row.append(item.get(col, "---"))
            self.c.addrow(cvalues=row)
            
        for item in self.colcats:
            row = ["column"]
            for col in self.columns:
                row.append(item.get(col, "---"))
            self.c.addrow(cvalues=row)
            
        self.c.generate()

    
class NonProcPivotTable(object):
    """Accumulate an object that can be turned into a basic pivot table once a procedure state can be established"""
    
    def __init__(self, omssubtype, outlinetitle="", tabletitle="", caption="", rowdim="", coldim="", columnlabels=[],
                 procname="Messages"):
        """omssubtype is the OMS table subtype.
        caption is the table caption.
        tabletitle is the table title.
        columnlabels is a sequence of column labels.
        If columnlabels is empty, this is treated as a one-column table, and the rowlabels 
        are used as the values with the label column hidden
        
        procname is the procedure name.  It must not be translated."""
        
        attributesFromDict(locals())
        self.rowlabels = []
        self.columnvalues = []
        self.rowcount = 0

    def addrow(self, rowlabel=None, cvalues=None):
        """Append a row labelled rowlabel to the table and set value(s) from cvalues.
        
        rowlabel is a label for the stub.
        cvalues is a sequence of values with the same number of values are there are columns in the table."""
        
        if cvalues is None:
            cvalues = []
        self.rowcount += 1
        if rowlabel is None:
            self.rowlabels.append(str(self.rowcount))
        else:
            self.rowlabels.append(rowlabel)
        self.columnvalues.extend(cvalues)
        
    def generate(self):
        """Produce the table if it has any rows, assuming that a procedure state is now in effect or possible"""
        
        privateproc = False
        if self.rowcount > 0:
            try:
                table = spss.BasePivotTable(self.tabletitle, self.omssubtype)
            except:
                spss.EndDataStep()  # just in case there is a dangling DataStep
                spss.StartProcedure(self.procname)
                privateproc = True
                table = spss.BasePivotTable(self.tabletitle, self.omssubtype)
            if self.caption:
                table.Caption(self.caption)
            if self.columnlabels != []:
                table.SimplePivotTable(self.rowdim, self.rowlabels, self.coldim, 
                    self.columnlabels, self.columnvalues)
            else:
                table.Append(spss.Dimension.Place.row,"rowdim",hideName=True,hideLabels=True)
                table.Append(spss.Dimension.Place.column,"coldim",hideName=True,hideLabels=True)
                colcat = spss.CellText.String("Message")
                for r in self.rowlabels:
                    if isinstance(r, (int, float)):
                        r = str(r)
                    cellr = spss.CellText.String(r)
                    table[(cellr, colcat)] = cellr
            if privateproc:
                spss.EndProcedure()
                
def attributesFromDict(d):
    """build self attributes from a dictionary d."""
    self = d.pop('self')
    for name, value in d.iteritems():
        setattr(self, name, value)
def _isseq(obj):
    """Return True if obj is a sequence, i.e., is iterable.
    
    Will be False if obj is a string or basic data type"""
    
    # differs from operator.isSequenceType() in being False for a string
    
    if isinstance(obj, basestring):
        return False
    else:
        try:
            iter(obj)
        except:
            return False
        return True
